
#include "config.h"
#include "types.h"

#include "p0f-db.h"
#include "p0f-inline.h"

#ifdef ENABLE_POSTGRESQL
static PGconn *pg_conn;
unsigned long long pdb_pgchecksys(_u8 * a, _u8 * os, _u8 * desc);
#endif

#ifdef ENABLE_MYSQL
static MYSQL *mysql_conn;
unsigned long long pdb_mysqlchecksys(_u8 * a, _u8 * os, _u8 * desc);
#endif

#ifdef ENABLE_SQLITE
static sqlite3 *sqlite_conn;
static int row_count;
unsigned long long sqlite_sid;
unsigned long long pdb_sqlitechecksys(_u8 * a, _u8 * os, _u8 * desc);
#endif

#ifdef NO_STRNDUP
char *temp_strndup(const char *string, size_t n)
{
    char *newstring;
    size_t len = strlen(string);
    if (len > n)
	len = n;
    newstring = (char *) malloc(len + 1);
    if (newstring == NULL)
	return NULL;
    newstring[len] = '\0';
    return (char *) memcpy(newstring, string, len);
}

#define strndup temp_strndup
#endif

char *pdb_escape(char *str);
void pdb_parseconfig();

/*-----*/

#ifdef ENABLE_SQLITE

void
pdb_sqliteout(_u8 * a, _u16 sp, _u8 * os, _u8 * desc, struct timeval pts,
	      char *natAnswer, char *ecnAnswer, char *firewallAnswer,
	      char *lookupAnswer, _u16 distanceAnswer)
{

    int rc;
    char *zErrMsg = 0;
    char query[PDB_QUERY_LEN];
    char additionalQuery[PDB_QUERY_LEN];
    char *os_escaped;
    char *desc_escaped;
    unsigned long long sid;

    char tstamp[20];
    size_t i;
    struct tm tim;
    time_t now;

    if (!sqlite_conn) {
	/* debug("PDB_SL: Not connected to DB.\n"); */
	return;
    }

    os_escaped = pdb_escape(os);
    desc_escaped = pdb_escape(desc);
    sid = pdb_sqlitechecksys(a, os_escaped, desc_escaped);
    if (!sid) {	/* IP not in DB -- make a new record */
	now = time(NULL);
	tim = *(localtime(&now));
	i = strftime(tstamp, 20, "%Y-%m-%d %H:%M:%S", &tim);
	snprintf(query, PDB_QUERY_LEN,
		 "INSERT INTO system (ip_addr, name, first_tstamp, last_tstamp) VALUES (%u, '%.32s %.64s', '%s', '%s');",
		 (a[0] << 24) | (a[1] << 16) | (a[2] << 8) | a[3], os_escaped,
		 desc_escaped, tstamp, tstamp);
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "INSERT INTO system_details (sid, ip_addr, nat, ecn, firewall, lookup_link, distance) VALUES (last_insert_rowid(), %u, '%.32s', '%.32s', '%.32s', '%.64s', '%d');",
		 (a[0] << 24) | (a[1] << 16) | (a[2] << 8) | a[3],
		 natAnswer, ecnAnswer, firewallAnswer, lookupAnswer,
		 distanceAnswer);
    } else {			/* IP is already in DB -- update time stamp */
	now = time(NULL);
	tim = *(localtime(&now));
	i = strftime(tstamp, 30, "%Y-%m-%d %H:%M:%S", &tim);
	snprintf(query, PDB_QUERY_LEN,
		 "UPDATE system SET last_tstamp = '%s' WHERE sid = %llu;",
		 tstamp, sid);
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "UPDATE system_details SET nat='%.32s', ecn='%.32s', firewall='%.32s', lookup_link='%.64s', distance='%d' WHERE sid = %llu;",
		 natAnswer, ecnAnswer, firewallAnswer, lookupAnswer,
		 distanceAnswer, sid);
    }

    free(os_escaped);
    free(desc_escaped);

    /* No callback needed, as we use INSERT or UPDATE */
    /* FIXME: We should execute both queries in one transaction! */
    rc = sqlite3_exec(sqlite_conn, query, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
	debug("PDB_SL: Insert/Update failed: %s\n", zErrMsg);
	sqlite3_free(zErrMsg);
	sqlite3_close(sqlite_conn);
	fatal("PDB_SL: Exiting.\n");
    }
    rc = sqlite3_exec(sqlite_conn, additionalQuery, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
	debug("PDB_SL: Insert/Update failed: %s\n", zErrMsg);
	sqlite3_free(zErrMsg);
	sqlite3_close(sqlite_conn);
	fatal("PDB_SL: Exiting.\n");
    }
}

/*-----*/

void pdb_sqliteconnect(const char *conninfo)
{

#define MCFLEN 128

    char *db;

    char *myvar;
    char *myval;
    char *mycon;

    int rc;

    void setconnopt(const char *var, const char *val) {
	if (strncasecmp(var, "DBNAME", MCFLEN) == 0) {	//  database name
	    db = strcat(strndup(val, MCFLEN), ".db");
	    if (!db)
		fatal("PDB_SL: Out of memory.\n");
	} else {		// unknown keyword in SQLite connection string
	    fatal("PDB_SL: Unknown keyword '%s' in config file.\n", var);
	}
    }

    // parse conninfo into separate values
    mycon = strndup(conninfo, PDB_CONF_LINE_LEN);
    if (!mycon)
	fatal("PDB_SL: Out of memory.\n");
    myvar = strtok(mycon, "=");
    myval = strtok(NULL, " ");
    if (myvar && myval)
	setconnopt(myvar, myval);
    while (myvar && myval) {	// if there're any more arguments, get them
	myvar = strtok(NULL, "=");
	if (!myvar)
	    break;
	myval = strtok(NULL, " ");
	if (!myval)
	    break;
	setconnopt(myvar, myval);
    }

    // do connect!
    rc = sqlite3_open(db, &sqlite_conn);
    if (rc) {
	debug("PDB_SL: Can't open database: %s\n",
	      sqlite3_errmsg(sqlite_conn));
	sqlite3_close(sqlite_conn);
	fatal("PDB_SL: Exiting.\n");
    } else
	debug("PDB_SL: Connected to '%s' database.\n", db);
}

/*-----*/

void pdb_sqlitedisconnect()
{

    if (sqlite_conn) {
	debug("PDB_SL: Closing connection to SQLite DB.\n");
	sqlite3_close(sqlite_conn);
    }
}

/*-----*/

static int
pdb_sqlitechecksys_callback(void *NotUsed, int argc, char **argv,
			    char **azColName)
{

    if (argc != 0) {
	row_count++;
        sqlite_sid = atoi(argv[0]);
    }
    return 0;
}

/*-----*/

unsigned long long pdb_sqlitechecksys(_u8 * a, _u8 * os, _u8 * desc)
{

    char ipquery[PDB_QUERY_LEN];
    char *zErrMsg = 0;
    int rc;

    snprintf(ipquery, PDB_QUERY_LEN,
	     "SELECT sid, ip_addr FROM system WHERE ip_addr = %u AND name = '%.32s %.64s' LIMIT 1;",
	     (a[0] << 24) | (a[1] << 16) | (a[2] << 8) | a[3], os, desc);

    row_count = 0;		//reset row counter
    sqlite_sid = 0;
    rc = sqlite3_exec(sqlite_conn, ipquery, pdb_sqlitechecksys_callback, 0,
		      &zErrMsg);
    if (rc != SQLITE_OK) {
	debug("PDB_SL: Check IP failed: %s\n", zErrMsg);
	sqlite3_free(zErrMsg);
	sqlite3_close(sqlite_conn);
	fatal("PDB_SL: Exiting due to DB failure.\n");
    }

    return sqlite_sid;
}

#endif				/* ifdef ENABLE_SQLITE */

/*-----*/

#ifdef ENABLE_POSTGRESQL

void
pdb_pgout(_u8 * a, _u16 sp, _u8 * os, _u8 * desc, struct timeval pts,
	  char *natAnswer, char *ecnAnswer, char *firewallAnswer,
	  char *lookupAnswer, _u16 distanceAnswer)
{
    char query[PDB_QUERY_LEN];
    char additionalQuery[PDB_QUERY_LEN];
    PGresult *res;
    char *os_escaped;
    char *desc_escaped;
    unsigned long long sid;

    if (!pg_conn) {
	/* debug("PDB_PG: Not connected to DB.\n"); */
	return;
    }

    /* Start a transaction block */
    res = PQexec(pg_conn, "BEGIN");
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	debug("PDB_PG: BEGIN command failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    }
    PQclear(res);

    /* do insert/update */

    os_escaped = pdb_escape(os);
    desc_escaped = pdb_escape(desc);
    sid = pdb_pgchecksys(a, os_escaped, desc_escaped);
    if (!sid) {		/* IP not in DB -- make a new record */
	snprintf(query, PDB_QUERY_LEN,
		 "INSERT INTO system (ip_addr, name, first_tstamp, last_tstamp) VALUES ('%d.%d.%d.%d', '%.32s %.64s', '%.32s', '%.32s');",
		 a[0], a[1], a[2], a[3], os_escaped, desc_escaped,
		 asctime(localtime(&pts.tv_sec)),
		 asctime(localtime(&pts.tv_sec)));
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "INSERT INTO system_details (sid, ip_addr, nat, ecn, firewall, lookup_link, distance) VALUES (lastval(), '%d.%d.%d.%d', '%.32s', '%.32s', '%.32s', '%.64s', '%d');",
		 a[0], a[1], a[2], a[3], natAnswer, ecnAnswer,
		 firewallAnswer, lookupAnswer, distanceAnswer);
    } else {			/* IP is already in DB -- update time stamp */
	snprintf(query, PDB_QUERY_LEN,
		 "UPDATE system SET last_tstamp = '%.32s' WHERE sid = '%llu';",
		 asctime(localtime(&pts.tv_sec)), sid);
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "UPDATE system_details SET nat='%.32s', ecn='%.32s', firewall='%.32s', lookup_link='%.64s', distance='%d' WHERE sid = '%llu';",
		 natAnswer, ecnAnswer, firewallAnswer, lookupAnswer, distanceAnswer, sid);
    }
    free(os_escaped);
    free(desc_escaped);

    res = PQexec(pg_conn, query);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	debug("PDB_PG: Insert/Update failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    }
    PQclear(res);

    res = PQexec(pg_conn, additionalQuery);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	debug("PDB_PG: Insert/Update failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    }
    PQclear(res);

    /* end the transaction */
    res = PQexec(pg_conn, "END");
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	debug("PDB_PG: END command failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    }
    PQclear(res);
}

/*-----*/

void pdb_pgconnect(const char *conninfo)
{
    char query[PDB_QUERY_LEN];
    PGresult *res;
    int nTuples;

    pg_conn = PQconnectdb(conninfo);
    if (PQstatus(pg_conn) != CONNECTION_OK) {
	debug("PDB_PG: Connection to database '%s' failed.\n",
	      PQdb(pg_conn));
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    } else
	debug("PDB_PG: Connected to '%s' database.\n", PQdb(pg_conn));

    // check p0f db scheme version
    snprintf(query, PDB_QUERY_LEN, "SELECT version FROM scheme WHERE version ='%d';", PDB_SCHEME_VERSION);
    res = PQexec(pg_conn, query);
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
        debug("PDB_PG: scheme version check failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting due to DB failure.\n");
    }
    nTuples = PQntuples(res);
    PQclear(res);
    // if the query returned exactly 1 row, then the scheme version is correct
    if (nTuples != 1) {
        debug("PDB_PG: The underlying database scheme is incorrect.\nThis p0f-db version supports scheme #%d.\n", PDB_SCHEME_VERSION);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting.\n");
    }

}

/*-----*/

void pdb_pgdisconnect()
{
    if (pg_conn) {
	debug("PDB_PG: Closing connection to '%s'.\n", PQdb(pg_conn));
	PQfinish(pg_conn);
    }
}

/*-----*/

unsigned long long pdb_pgchecksys(_u8 * a, _u8 * os, _u8 * desc)
{
    char ipquery[PDB_QUERY_LEN];
    PGresult *res;
    int nTuples;
    long sid;

    snprintf(ipquery, PDB_QUERY_LEN,
	     "SELECT sid, ip_addr FROM system WHERE ip_addr = '%d.%d.%d.%d' AND name = '%.32s %.64s' LIMIT 1;",
	     a[0], a[1], a[2], a[3], os, desc);
    res = PQexec(pg_conn, ipquery);
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
	debug("PDB_PG: Check IP failed: %s", PQerrorMessage(pg_conn));
	PQclear(res);
	PQfinish(pg_conn);
	fatal("PDB_PG: Exiting due to DB failure.\n");
    }

    /* Return the number of rows selected */
    nTuples = PQntuples(res);
    sid = 0;
    if (nTuples)
        sid = atoi(PQgetvalue(res, 0, 0));
    PQclear(res);

    return sid;
}

#endif				/* ifdef ENABLE_POSTGRESQL */

/*-----*/

#ifdef ENABLE_MYSQL

void
pdb_mysqlout(_u8 * a, _u16 sp, _u8 * os, _u8 * desc, struct timeval pts,
	     char *natAnswer, char *ecnAnswer, char *firewallAnswer,
	     char *lookupAnswer, _u16 distanceAnswer)
{
    char query[PDB_QUERY_LEN];
    char additionalQuery[PDB_QUERY_LEN];
    char *os_escaped;
    char *desc_escaped;
    my_ulonglong sid;

    if (!mysql_conn) {
	/* debug("PDB_MY: Not connected to DB.\n"); */
	return;
    }

    os_escaped = pdb_escape(os);
    desc_escaped = pdb_escape(desc);
    sid = pdb_mysqlchecksys(a, os_escaped, desc_escaped);
    if (!sid) {	/* IP not in DB -- make a new record */
	snprintf(query, PDB_QUERY_LEN,
		 "INSERT INTO system (ip_addr, name, first_tstamp, last_tstamp) VALUES (INET_ATON('%d.%d.%d.%d'), '%.32s %.64s', FROM_UNIXTIME(%.32ld), FROM_UNIXTIME(%.32ld));",
		 a[0], a[1], a[2], a[3], os_escaped, desc_escaped,
		 mktime(localtime(&pts.tv_sec)),
		 mktime(localtime(&pts.tv_sec)));
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "INSERT INTO system_details (sid, ip_addr, nat, ecn, firewall, lookup_link, distance) VALUES (LAST_INSERT_ID(), INET_ATON('%d.%d.%d.%d'), '%.32s', '%.32s', '%.32s', '%.64s', '%d');",
		 a[0], a[1], a[2], a[3], natAnswer, ecnAnswer, firewallAnswer, lookupAnswer, distanceAnswer);
    } else {			/* IP is already in DB -- update time stamp */
	snprintf(query, PDB_QUERY_LEN,
		 "UPDATE system SET last_tstamp = FROM_UNIXTIME(%.32d) WHERE sid = '%llu'",
		 mktime(localtime(&pts.tv_sec)), sid);
	snprintf(additionalQuery, PDB_QUERY_LEN,
		 "UPDATE system_details SET nat='%.32s', ecn='%.32s', firewall='%.32s', lookup_link='%.64s', distance='%d' WHERE sid = '%llu'",
		 natAnswer, ecnAnswer, firewallAnswer, lookupAnswer, distanceAnswer, sid);
    }
    free(os_escaped);
    free(desc_escaped);

    if (mysql_query(mysql_conn, query)) {	/* If this returns non zero -> error occured */
	debug("PDB_MY: Insert/Update failed: %s", mysql_error(mysql_conn));
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting.\n");
    }
    if (mysql_query(mysql_conn, additionalQuery)) {	/* If this returns non zero -> error occured */
	debug("PDB_MY: Insert/Update failed: %s", mysql_error(mysql_conn));
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting.\n");
    }
}

/*-----*/

void pdb_mysqlconnect(const char *conninfo)
{
#define MCFLEN 128

    char *host = "localhost";
    char *user;
    char *password;
    char *db;
    int port = 3306;

    char *myvar;
    char *myval;
    char *mycon;

    char query[PDB_QUERY_LEN];
    MYSQL_RES *res;
    my_ulonglong rows;

    void setconnopt(const char *var, const char *val) {
	if (strncasecmp(var, "HOST", MCFLEN) == 0) {	/* host part */
	    host = strndup(val, MCFLEN);
	    if (!host)
		fatal("PDB_MY: Out of memory.\n");
	} else if (strncasecmp(var, "PORT", MCFLEN) == 0) {	/* port # */
	    port = atol(val);
	} else if (strncasecmp(var, "USER", MCFLEN) == 0) {	/* user name */
	    user = strndup(val, MCFLEN);
	    if (!user)
		fatal("PDB_MY: Out of memory.\n");
	} else if (strncasecmp(var, "PASSWORD", MCFLEN) == 0) {	/* user password */
	    password = strndup(val, MCFLEN);
	    if (!password)
		fatal("PDB_MY: Out of memory.\n");
	} else if (strncasecmp(var, "DBNAME", MCFLEN) == 0) {	/* database name */
	    db = strndup(val, MCFLEN);
	    if (!db)
		fatal("PDB_MY: Out of memory.\n");
	} else {		/* unknown keyword in MySQL connection string */

	    fatal("PDB_MY: Unknown keyword '%s' in config file.\n", var);
	}
    }

    // parse conninfo into separate values
    mycon = strndup(conninfo, PDB_CONF_LINE_LEN);
    if (!mycon)
	fatal("PDB_MY: Out of memory.\n");
    myvar = strtok(mycon, "=");
    myval = strtok(NULL, " ");
    if (myvar && myval)
	setconnopt(myvar, myval);
    while (myvar && myval) {	/* if there're any more arguments, get them */
	myvar = strtok(NULL, "=");
	if (!myvar)
	    break;
	myval = strtok(NULL, " ");
	if (!myval)
	    break;
	setconnopt(myvar, myval);
    }
    // do connect!
    mysql_conn = mysql_init(NULL);
    if (!mysql_conn)
	fatal("PDB_MY: Out of memory.\n");
    mysql_real_connect(mysql_conn, host, user, password, db, port, NULL,
		       0);
    if (!mysql_conn) {		/* connection failed? */
	debug("PDB_MY: Connection to database '%s' failed.\n",
	      mysql_error(mysql_conn));
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting.\n");
    } else
	debug("PDB_MY: Connected to '%s' database.\n", db);

    // check p0f db scheme version
    snprintf(query, PDB_QUERY_LEN, "SELECT version FROM scheme WHERE version = '%d';", PDB_SCHEME_VERSION);
    if (mysql_query(mysql_conn, query)) {
        debug("PDB_MY: scheme version check failed: %s", mysql_error(mysql_conn));
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting due to DB failure.\n");
    }
    res = mysql_store_result(mysql_conn);
    rows = mysql_num_rows(res);
    mysql_free_result(res);
    // if the query returned exactly 1 row, then the scheme version is correct
    if (rows != 1) {
        debug("PDB_MY: The underlying database scheme is incorrect.\nThis p0f-db version supports scheme #%d.\n", PDB_SCHEME_VERSION);
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting.\n");
    }

}

/*-----*/

void pdb_mysqldisconnect()
{
    if (mysql_conn) {
	debug("PDB_MY: Closing connection to MySQL DB.\n");
	mysql_close(mysql_conn);
    }
}

/*-----*/

unsigned long long pdb_mysqlchecksys(_u8 * a, _u8 * os, _u8 * desc)
{
    char ipquery[PDB_QUERY_LEN];
    MYSQL_RES *res;
    my_ulonglong rows;
    my_ulonglong sid;
    MYSQL_ROW row;

    snprintf(ipquery, PDB_QUERY_LEN,
	     "SELECT sid, ip_addr FROM system WHERE ip_addr = INET_ATON('%d.%d.%d.%d') AND name = '%.32s %.64s' LIMIT 1;",
	     a[0], a[1], a[2], a[3], os, desc);

    if (mysql_query(mysql_conn, ipquery)) {
	debug("PDB_MY: Check IP failed: %s", mysql_error(mysql_conn));
	mysql_close(mysql_conn);
	fatal("PDB_MY: Exiting due to DB failure.\n");
    }

    res = mysql_store_result(mysql_conn);
    rows = mysql_num_rows(res);
    sid = 0;
    if (rows) {
        row = mysql_fetch_row(res);
	sid = atoi(row[0]);
    }
    mysql_free_result(res);

    return sid;
}

#endif				/* ifdef ENABLE_MYSQL */

/*-----*/

char *pdb_escape(char *str)
{
/*  Ideas borrowed from SNORT (www.snort.org) 
 *  database output plugin.
 */
    char *buf, *buf_start;

    buf = (char *) malloc(strlen(str) * 2 + 1);
    memset(buf, 0, strlen(str) * 2 + 1);
    buf_start = buf;

    while (*str) {
	switch (*str) {
	case '\'':
	    *buf = '\'';
	    buf++;
	    *buf = '\'';
	default:
	    *buf = *str;
	}
	str++;
	buf++;
    }

    return buf_start;
}

/*-----*/

void pdb_parseconfig()
{
    int linenum;
    FILE *conff;
    char confline[PDB_CONF_LINE_LEN];
    char *conftok;

    conff = fopen(PDB_CONF_FILE_NAME, "r");
    if (!conff) {
	fatal("PDB: Cannot open config file: %s\n", PDB_CONF_FILE_NAME);
    }

    linenum = 0;
    while (fgets(confline, PDB_CONF_LINE_LEN, conff)) {
	linenum++;

	conftok = strtok(confline, " ");	/* get a first token */
	if (conftok && conftok[0] != '#' && conftok[0] != '\n') {	/* if this is comment/empty line -- throw it off */
	    if (strcasecmp(conftok, "POSTGRESQL") == 0) {	/* postgres output? */
#ifndef ENABLE_POSTGRESQL
		fatal("PDB: %s: PostgreSQL support not compiled.\n",
		      PDB_CONF_FILE_NAME);
#else
		conftok = strtok(NULL, "\n");
		/* conftok now should be equal to connection info (user, dbname, etc.) */
		pdb_pgconnect(conftok);
#endif
	    } else if (strcasecmp(conftok, "MYSQL") == 0) {	/* mysql output? */
#ifndef ENABLE_MYSQL
		fatal("PDB: %s: MySQL support not compiled.\n",
		      PDB_CONF_FILE_NAME);
#else
		conftok = strtok(NULL, "\n");
		/* conftok now should be equal to connection info (user, dbname, etc.) */
		pdb_mysqlconnect(conftok);
#endif
	    } else if (strcasecmp(conftok, "SQLITE") == 0) {	/* sqlite output? */
#ifndef ENABLE_SQLITE
		fatal("PDB: %s: SQLite support not compiled.\n",
		      PDB_CONF_FILE_NAME);
#else
		conftok = strtok(NULL, "\n");
		/* conftok now should be equal to connection info (dbname!) */
		pdb_sqliteconnect(conftok);
#endif
	    } else {
		debug("PDB: Uknown keyword found\n");
		fatal("PDB: %s:%d: %s", PDB_CONF_FILE_NAME, linenum,
		      conftok);
	    }
	}			/* if (conftok && conftok[0] != '#') */
    }

    fclose(conff);
}

/*-----*/
